#    common.tcl
#
#   useful procedures

#    to dummy out for 7.5 on
proc register-interest args {}

set debug 0

proc first p {lindex $p 0}
proc second p {lindex $p 1}
proc third p {lindex $p 2}
proc fourth p {lindex $p 3}
proc fifth p {lindex $p 4}
proc sixth p {lindex $p 5}
proc seventh p {lindex $p 6}
proc eigth p {lindex $p 7}
proc eighth p {lindex $p 7}
proc nineth p {lindex $p 8}
proc ninth p {lindex $p 8}
proc tenth p {lindex $p 9}

proc last p {lindex $p [expr [llength $p]-1]}
proc last_but_one p {lindex $p [expr [llength $p]-2]}
proc last_but_two p {lindex $p [expr [llength $p]-3]}
proc last_but_three p {lindex $p [expr [llength $p]-4]}

proc front p {lreplace $p [expr [llength $p]-1] [expr [llength $p]-1]}

proc head p {lindex $p 0}
proc tail p {lreplace $p 0 0}
proc tail_from_second p {lreplace $p 0 0}
proc tail_from_third p {lreplace $p 0 1}
proc tail_from_fourth p {lreplace $p 0 2}

# Used for speeding up lreplace in-place. See http://wiki.tcl.tk/348
# Only seems to be consistently effective on arrays longer than ca. 20/50/100... i.e. not clear on min length :-(
# Syntax is: set theList [lreplace [K $theList [set theList {}]] first last ?element element ...?]
# N.B. 1) tcl 8.5 can use: set theList [lreplace $theList[set theList {}] first last ?element element ...?]
#         (concatenate theList with an empty theList, no need for K. Does seem faster.
#      2) tcl 8.5 has a new lset command
proc K {x y} {set x}

if {[info tclversion] != 7.4} then {proc transparent args {return $args}}

if {[info tclversion] == 7.4} then {
  rename file xfile
  proc file args {
     if {[first $args] != "join"} then {
        set m [eval xfile $args]
        return $m
     }
     set path [second $args]
     foreach item [lreplace $args 0 1] {append path /$item}
     return $path
  }
}

proc quote-strings l {
  set ll ""
  foreach f $l {append ll " " "$f"}
  return $ll
}

proc quote-list l {
   set ll {}
   foreach f $l {lappend ll [format "%s " $f]}
   return $ll
}

proc quote-text s {return "$s"}

proc make-list l {
   set ll {}
   foreach f $l {lappend ll [first $f]}
   return $ll
}

proc prefix-list {l p} {
   set ll {}
   foreach f $l {lappend ll $p$f}
   return $ll
}

proc list-take {as is} {
   set ll {}
   foreach e $is {lappend ll [lindex $as $e]}
   return $ll
}


proc file-contents {file} {
   set z [catch {open $file} f]
   if {$z != 0} then {midas-warning $f; return ""}
   set text [read $f]
   close $f
   return "$text"
}


proc pluralise-spectrum {n} {if {$n==1} then {return "spectrum"} else {return "spectra"}}

proc pluralise-word-y {w n} {if {$n==1} {return $w} {return [string trimright $w y]ies}}

proc pluralise-word {w n} {
   if {[string range $w e e]=="y"} then {
      return [pluralise-word-y $w $n]
   } else {if {$w=="spectrum"} then {
      return [pluralise-spectrum $n]
   } else {
      if {$n==1} {return $w} {return [set w]s}
   }}
}

proc filetolist {str} {
#    converts stdout or stderr output to the form of a list with one output
#    line per list item

  set list {}
  set s $str
  set j 0
  while {$j == 0} {
    set k [string first \n $s]
    if {$k < 0} then {
      lappend list [string range $s 0 end]
      return $list
    }
    lappend list [string range $s 0 [expr $k-1]]
    incr k
    set s [string range $s $k end]
 }
}

proc scan-num s {
   if {[string first "0x" $s]==0} then {
      scan [string range $s 2 end] %x zzz; return $zzz}
   scan $s %f zzz; return $zzz
}

proc scan-int s {
   if {[string first "0x" $s]==0} then {
      scan [string range $s 2 end] %x zzz; return $zzz}
   scan $s %d zzz; return $zzz
}

proc scan-int s {
# note return of ??? if s fails to scan
   if {[string first "0x" $s]==0} then {
     set z [scan [string range $s 2 end] %x zzz]
   } else {
     set z [scan $s %d zzz]
   }
   if {$z==1} then {return $zzz} else {return "???"}
}

proc scan-num s {
# note return of ??? if s fails to scan
   if {[string first "0x" $s]==0} then {
     set z [scan [string range $s 2 end] %x zzz]
   } else {
     set z [scan $s %f zzz]
   }
   if {$z==1} then {return $zzz} else {return "???"}
}

proc truncate r {
   set i [string first "." $r]
   if {$i<0} then {return $r}
   incr i -1
   return [string range $r 0 $i]
}

proc strip-to-dot s {string range $s [expr 1+[string first . $s]] end }

proc strip-dots l {
   set ll {}
   foreach n $l {if {$n!="." && $n!=".."} then {lappend ll $n}}
   return $ll
}

proc drop-trailing-nulls ss {
   set n [llength $ss]
   if {$n==0} then {return {}}
   incr n -1
   if {[lindex $ss $n]=={}} then {drop-trailing-nulls [lreplace $ss $n $n]} \
   else {return $ss}
}

proc drop-trailing-spaces ss {
   set n [llength $ss]
   if {$n==0} then {return {}}
   incr n -1
   if {[lindex $ss $n]==" "} then {drop-trailing-spaces [lreplace $ss $n $n]} \
   else {return $ss}
}

proc DeleteTask args {
  if {[info tclversion] != 7.4} then {
       #    assume 7.5 or later

     if {[first $args] == ""} then {set task [Get_My_Name]} \
        else {set task [first $args]}

     delete_task $task
  } else {
    set task [Get_My_Name]
    destroy .
  }
  puts stdout "Deletetask: $task"
}

proc invert {b n} {
   set i [expr int(pow (2, $b - 1))]
   set r 0
   while {$n!=0} {
      if {$n%2} then {incr r $i}
      set n [expr $n/2]
      set i [expr $i/2]
   }
   return $r
}

proc bits-in n {
   set i 0
   set ll {}
   while {$n!=0} {
      if {$n%2} then {lappend ll $i}
      set n [expr $n/2]
      incr i
   }
   return $ll
}

#    system log interface

proc insert-log s {

     set z [catch {inform sl insert [transparent $s]} m]
     if {$z != 0} then {
         puts stdout "$m"
         puts stdout "$s"
     }
}


#    directory access support procedures

set directorylist {}
set directory_list_option "l"

proc read-directory-names {directory} {
   global directorylist directory_list_option

   set z [catch {split [:list_directory -$directory_list_option $directory] \n} m]
   if {$z != 0} then {
        display-midas-error-report self "" "$m" "\n" "Attempting to read directory" $directory
        clear-footer
        error $m
   }

   set l {}
   foreach line $m {
      set x [first [split $line ""]]
      if {$x == "d" || $x == "-"} then {lappend l [last $line]}
   }
   set directorylist $l
   return $directorylist
}

proc read-directory-full {directory} {
   global directorylist directory_list_option

   set z [catch {
      set type [file type $directory]
      tail [split [:list_directory -$directory_list_option [  \
         switch $type {
            directory {set directory}
            link      {file readlink $directory}
            default   {error "$directory is not a directory!!!"}
         }
      ]] \n]} m]

   if {$z != 0} then {
        display-midas-error-report self "" "$m"
        clear-footer
        error $m
   }

   set directorylist $m
   return $directorylist
}

proc read-directory-default {directory} {
   global directorylist

   set z [catch {eg "(directory :path '$directory' ?files)"} m]
   if {$z != 0} then {
        display-midas-error-report self $m "Attempting to read directory" $directory
        clear-footer
        error $m
   }
   set directorylist [lsort $m]
   return $directorylist
}

proc read-directory {directory} {read-directory-full $directory}

proc directory-down {directory selection} {
   global directorylist env

   if {[llength $selection]==0} then {
      set-footer "no directory selected!!!"
      return $directory
   }
   if {[llength $selection]>1} then {
      set-footer "multiple selections!!!"
      return $directory
   }

   set target [lindex $directorylist $selection]
   set filename [last $target]

   switch $env(platform) {
      unix    {
         if {[string index $target 0] == "l"} then {
#    target is a symlink -  first need to look at the file
             set filename [last_but_two $target]
             if {![file isdirectory [file join $directory $filename]]} then {
                set-footer "$filename is not a directory"
                return $directory
             }
             return [file readlink [file join $directory $filename]]
         }
         if {![file isdirectory [file join $directory $filename]]} then {
            set-footer "$filename is not a directory"
            return $directory
         }
      }
      windows {
         if {![file isdirectory [file join $directory $filename]]} then {
            set-footer "$filename is not a directory"
            return $directory
         }
      }
      default {puts stderr "in directory-down $env(platform) not supported"; return ""}
   }
   if {$filename == "."} then  {return $directory}
   if {$filename == ".."} then {return [directory-up $directory]}
   return [file join $directory $filename]
}

proc directory-up {directory} {

   set i [string last / $directory]
   if {$i<0} then {return $directory}
   if {$i==0} then {set directory "/"}
   if {$i>0} then {set directory [string range $directory 0 [expr $i-1]]}

   return $directory
}

proc paint-directory {directory} {
   global directorylist

   :modify DIRECTORY     :value "$directory"
   :modify DIRECTORYLIST :empty T :strings $directorylist :see 0
   set-footer "[llength $directorylist] files/directories"
}

#    directory browser functions

set browserdirectory $env(HOME)
set browserfile ""

proc midas-browser {mode command args}  {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file
   global env

  if {[info tclversion] < "8.0"} then {

     set frame [self frame-name].browser
     :frame $frame :label "Directory Browser for [self frame-title]" :show-footer T :transient T :resize F

     :panel BROWSER :padx 10 :pady 10 :col-gap 10 :row-gap 10
     :layout H
     :text DIRECTORY  :w 60 :shifters T  :value "$directory"  :command browser-directory
     :button BROWSER-UP :label Up :command browser-up
     :next-row T
     :list DIRECTORYLIST :exclusive T :rows 10 :cols 80  :command browser-selection
     :next-row T
     if {$mode == "W" || $mode == "w"} then {
        :text BROWSER-FILE :label Name :w 60 :value "$file" :command browser-file
        :button BROWSER-OK :label Ok :command "browser-delete; $command $args"
        :button BROWSER-MKDIR :label "Make Directory"  :command browser-mkdir
     } else {
        :text BROWSER-FILE :label File :w 60 :value "$file" :command browser-file
        :hskip 50
        :button BROWSER-OK :label Ok :command "browser-delete; $command $args"
     }

     :show T

     browser-redisplay

   } else {

#    for tk8.0 amd later use the built in browser function
#    This uses the canvas function and requires a pale coloured
#    background

     option add *Canvas.background  #ffffff
     if {$mode == "W" || $mode == "w"} then {
        set r [tk_getSaveFile -parent [self frame-name] \
              -title "Directory Browser for [self frame-title]" \
              -initialdir $directory  -initialfile $file]
     } else {
        set r [tk_getOpenFile -parent [self frame-name] \
              -title "Directory Browser for [self frame-title]" \
              -initialdir $directory  -initialfile $file]
     }
     option add *Canvas.background  $env(tkcanvasbackground)
     if {$r != ""} then {
        set directory [file dirname $r]
        set file [file tail $r]
        $command $args
     }
   }
}

proc browser-delete {} {destroy-frame [self frame-name].browser}

proc browser-redisplay {} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file
   global directorylist

   read-directory $directory
   :modify DIRECTORY     :value "$directory"
   :modify DIRECTORYLIST :empty T :strings $directorylist :see 0
   set-footer [self frame-name].browser "[llength $directorylist] files/directories"

   :modify BROWSER-FILE :value "$file"
}

proc browser-directory {w v} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file

   set directory $v
   set file ""
   browser-redisplay
}

proc browser-file {w v} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file

   set file "$v"
   :modify BROWSER-FILE :value "$file"
}

proc browser-mkdir {} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file

   set z [catch {:file mkdir $directory/$file} m]
   if {$z != 0} then {midas-warning "Unable to create directory $directory/$file\n$m"}

   set file ""
   browser-redisplay
}

proc browser-up {} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file

  set directory [directory-up $directory]
  browser-redisplay
}

proc browser-selection {w target i} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file

   set name [last $target]
   if {$name == "." || $name == ".."} then  {return}
   if {[file isdirectory $directory/$name]} then {
      set directory $directory/$name
      set file ""
      browser-redisplay
   } else {
      set file $name
      :modify BROWSER-FILE :value $file
   }
}

#    checks for existance of system temp filing directory
#    if exists creates a unique directory within it and makes
#    this the current directory

proc midas-check-for-temp {} {
  global env
   if {[file exists $env(MIDASTempDir)] == 0} then {
      midas-warning "Temporary file directory $env(MIDASTempDir) does not exist\n\
        You can set the environment variable MIDASTempDir to define an alternative location"
      return 0
   } else {
      catch {:file mkdir [file join $env(MIDASTempDir) tcl[pid]]}
      cd [file join $env(MIDASTempDir) tcl[pid]]
      return 1
   }
}

#    utility procs for handling of eg options

proc get-eg-option k {
   global env
   foreach p $env(EGOPTIONS) {
      if {[first $p]==$k} then {return [second $p]}
   }
   error "unknown option: $k"
}

proc set-eg-option {k v} {
   global env
   set ll {}
   foreach p $env(EGOPTIONS) {
      if {[first $p]!=$k} then {lappend ll $p}
   }
   lappend ll [list $k $v]
   set env(EGOPTIONS) $ll
}

proc edit-defaults {key stuff} {
   global debug
   if $debug then {puts stdout $key; puts stdout $stuff}
   set lines [load-defaults]
   set range [find-keyword $lines $key]
   if {[llength $range]==0} then \
      {save-defaults [linsert $lines [llength $lines] "## $key\n$stuff"]}\
   else \
      {save-defaults [lreplace $lines [first $range] [second $range] "## $key\n$stuff"]}
}

proc load-defaults {} {
   global env
   set z [catch {set f [open ~/.eg-defaults.tcl r]}]
   if $z!=0 then {return [list "# eg-session defaults for user $env(USER)" ""]}
   set lines [split [read $f] "\n"]
   close $f
   return $lines
}

proc save-defaults {lines} {
   set f [open ~/.eg-defaults.tcl w]
   puts $f [join $lines "\n"]
   close $f
}

proc find-keyword {lines key} {
   set i 0; set found 0
   foreach l $lines {
      switch $found {
      0 {if {[string first "## $key" $l]==0} then {set start $i; set found 1}}
      1 {if {[string first "##" $l]==0} then {set finish $i; set found 2}}
      2 {break}
      }
     incr i
   }
   if $found==0 then {return {}}
   if $found==1 then {return [list $start [expr $i-1]]}
   if $found==2 then {return [list $start [expr $finish-1]]}
}

#    the previous procs for eg-options are replaced by a standard setup of services for handling preferences

set preferences(preferences.filename)  "[file join $env(HOME) .midas preferences]"

proc get_preference_file {} {
    global env

#    look for the preference file 
    set pd [file join $env(HOME) .midas]
    if {![file exists $pd]} then {
         set z [catch {file mkdir $pd} m]
         if {$z != 0} then {insert-log "Unable to create directory $pd - $m"}
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd preferences]
    set z [catch {set f [open $pf r]} m]
    if {$z != 0} then {
        set z [catch {set f [open $pf w]} m]
        if {$z != 0} then {
            insert-log "Unable to access $pf - $m"
            return 1
        }
        close $f
        return {}
    }

#    and read it
    set lines [split [read $f] "\n"]
    close $f

    return $lines
}

proc put_preference_file {lines} {
   global env

#    look for the preference file 
    set pd [file join $env(HOME) .midas]
    if {![file exists $pd]} then {
         set z [catch {file mkdir $pd} m]
         if {$z != 0} then {insert-log "Unable to create directory $pd - $m"}
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd preferences]
    set z [catch {set f [open $pf w]} m]
    if {$z != 0} then {
        insert-log "Unable to access $pf - $m"
        return 1
    }

    foreach item $lines {
      if {$item != ""} {puts $f $item}
    }

    close $f
    return 0
}

#    save all preferences for the current task
proc update_preferences {} {
   global env preferences

   set lines [get_preference_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]
   if {[llength $range] > 0} then {
       set lines [lreplace $lines [first $range] [expr [second $range] + 1]]
   }

   lappend lines "## $key"
   foreach item [array names preferences] {
       lappend lines "$item $preferences($item)"
   }
   lappend lines "##"

   set z [put_preference_file $lines]
   return $z
}

#    load all preferences for the current task
proc load_preferences {} {
   global env preferences

   set lines [get_preference_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           set preferences([head $item]) [tail $item]
       }
   }
}

#    return task preferences
proc get_preferences {task} {
   global env

   set lines [get_preference_file]
   set key $task
   set range [find-keyword $lines $key]
   
   set preferences {}

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           lappend preferences [list [first $item] [second $item]]
       }
   } else {
       lappend preferences [list preferences.filename  "[file join $env(HOME) .midas preferences]"]
   }
   return $preferences
}

 
# ---utility procs for printing


proc lookup {k ps} {
   foreach p $ps {
      if {[first $p]==$k} then {return [second $p]}
   }
   error "keyword $k not in dictionary \{$ps\}"
}

proc get-print-template-specific {printer} {
   set z [catch {get-eg-option printernames} printers]
   if {$z != 0} then {return ""}
   foreach p $printers {
      if {[first $p]==$printer && [llength $p]>=3} then {return [third $p]}
   }
   return ""
}

proc get-print-template {printer} {
   set pt [get-print-template-specific $printer]
   if {$pt!=""} then {return $pt}
   set z [catch {get-eg-option printtemplate} pt]
   if {$z != 0} then {return ""}
   return $pt
}

proc substitute-print-template {template printer} {
   regsub %printer $template $printer ps
   return $ps
}

proc print-listing {stuff} {
    global env

      set printer [get-eg-option printer]
      set printernames [get-eg-option printernames]
      set printername [lookup $printer $printernames]
      set printdir [get-eg-option printtofiledir]
      set printcomm [substitute-print-template \
                    [get-print-template $printer] $printer]
      if {$printcomm==""} then {
         midas-warning "sorry, don't know how to print on printer \"$printer\""
         return
      }
      if {![midas-confirmation "Print text file on printer $printer"]} then {return}
      set-footer "printing on $printer..."

      switch $printer {
         file     {
             global print_stuff
             set print_stuff $stuff
             midas-browser W print-to-file
             return
         }
         default  {
            switch $env(platform) {
              unix    {set zz [catch {eval exec mp -o $stuff | $printcomm} m]}
              windows {set zz [catch {eval exec $printcomm $stuff} m]}
              default {puts stderr "text printing under $env(platform) not supported"}
            }
         }
      }
      if {$zz!=0} then {midas-warning "Printing to $printername failed \n $m"}
}

proc print-postscript {stuff} {
   global env

      set printer [get-eg-option printer]
      set printernames [get-eg-option printernames]
      set printername [lookup $printer $printernames]
      set printdir [get-eg-option printtofiledir]
      set printcomm [substitute-print-template \
                    [get-print-template $printer] $printer]
      if {$printcomm==""} then {
         midas-warning "sorry, don't know how to print on printer \"$printer\""
         return
      }
      if {![midas-confirmation "Print postscript file on printer $printer"]} then {return}
      set-footer "printing on $printer..."
      switch $printer {
         file     {
             global print_stuff
             set print_stuff $stuff
             midas-browser W print-to-file
             return
         }
         default  {
            switch $env(platform) {
              unix    {set zz [catch {eval exec cat $stuff | $printcomm} m]}
              windows {set zz [catch {eval exec $printcomm $stuff} m]}
              default {puts stderr "postscript printing under $env(platform) not supported"}
            }
         }
      }
      if {$zz!=0} then {midas-warning "Printing to $printername failed \n $m"}
}

proc print-to-file {args} {
   upvar #0 browserdirectory directory
   upvar #0 browserfile file
   global print_stuff

   set z [catch {:file rename $print_stuff $directory/$file} m]
   if {$z != 0} then {midas-warning "Unable to write $directory/$file \n $m"}
}

proc self op {
  global framename frame
  switch $op {
     frame-name {
       if {$framename != ""} then {set name $framename} else {set name $frame}
       if {[string match .* $name] == 0} then {set name .[string tolower $name]}
       return $name
     }
  }
}
#    self will be replaced by common.tk.tcl if/when it is loaded

proc Get_My_Name {} {global __interp; return [set __interp]}
set mytaskname [Get_My_Name]

proc inquire {v args} {
   global $v
   return [if {[llength $args]==0} then {
       set [set v]
   } else {
       set [set v]([first $args])
   }]
}

proc get-eg-option k {
   global env
   foreach p $env(EGOPTIONS) {
      if {[lindex $p 0]==$k} then {return [lindex $p 1]}
   }
   error "unknown option: $k"
}

proc loadsource file {
  global env tcl_platform
  if {[info tclversion] != 7.4} then {
     if {$tcl_platform(platform) == "unix"} then {set J ":"} else {set J ";"}
  } else {
     set J ":"
  }
  set path [split $env(TCLPATH) $J]
  set rc [catch {MIDASautoloadfile $file $path} m]
  if {$rc != 0} then {puts stderr "Error loading source in [info script]: $m"}
}

#    for debug output

proc insert-debug s {
  global env

    foreach i [Get_Task_Names] {
      if {$i == "rmtcntrl"} then {
        if {[info tclversion] != 7.4} then {
           inform rmtcntrl insert [string range [self frame-name] 1 end] [transparent $s]
        } else {
           send  $i insert [string range [self frame-name] 1 end] [transparent $s]
        }
        return
      }
    }
  puts stdout "$s"
}

#    interface routines to system dependent commands

proc date {args} {
  if {[info tclversion] != 7.4} then {
     switch $args {
        ""       {clock format [clock seconds] -format "%a %b %d %H:%M:%S %Z %Y"}
        default  {clock format [clock seconds] -format "[string range "$args" 1 end]"}
     }
  } else {
     eval exec date $args
  }
}

# Windows 95 no longer supported (needs command not cmd)
proc :list_directory {args} {
   global env

   switch $env(platform) {
      unix    {
         eval exec ls $args
      }
      windows {
         cd [last $args]
         exec cmd.exe /c dir
      }
      default {puts stderr "$env(platform) not supported"}
   }
}

proc :list_file {arg} {
   global env

   switch $env(platform) {
      unix    {
         eval exec cat $arg
      }
      windows {
         cd [file dirname $arg]
         exec cmd.exe /c type [file tail $arg]
      }
      default {puts stderr "$env(platform) not supported"}
   }
}

proc :file {cmd args} {
   global env
#    in general assume if tcl version == 7.4 then platform is unix

   switch $cmd {
      delete {
         if {[info tclversion] != 7.4} then {
             file delete -force [last $args]
         } else {
             eval exec rm $args
         }
      }
      rename {
         if {[info tclversion] != 7.4} then {
             file rename -force [last_but_one $args] [last $args]
         } else {
             eval exec mv $args
         }
      }
      copy {
         if {[info tclversion] != 7.4} then {
             file copy -force [first $args] [second $args]
         } else {
             eval exec cp -r $args
         }
      }
      chmod {
         if {[info tclversion] != 7.4} then {
#    to be supplied
             eval exec chmod  $args
         } else {
             eval exec chmod  $args
         }
      }
      mkdir {
         if {[info tclversion] != 7.4} then {
             file mkdir [last $args]
         } else {
             eval exec mkdir $args
             eval exec chmod 777 $args
         }
      }
   }
}



#    interface routines to the built-in eg command


proc :apparatus {cmd args} {
###############
###############

#    :apparatus  command  parameters
#        command  -  attach
#                    detach
#                    claim
#                    free

  set path ""; set name ""

  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

  switch $cmd {

      attach {
          set z [catch {eg "(apparatus :path '$path' :name '$name' !attach)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to attach apparatus file" $name
             error $m
          }
      }

      detach {
          set z [catch {eg "(apparatus :name '$name' !detach)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to detach apparatus file" $name
             error $m
          }
      }

      claim {
          set z [catch {eg "(apparatus :name '$name' !claim)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to claim apparatus" $name
             error $m
          }
      }

      free {
          set z [catch {eg "(apparatus :name '$name' !free)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to free apparatus" $name
             error $m
          }
      }

      read {
#          set z [catch {eg "(experiment ?apparatus)"} m]
#          set z [catch {exec tr "()" "{}" < $path/$name} m]

          set z [catch {open $path/$name} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read apparatus" $name
             error $m
          }
          set text [read $m]
          close $m

          set s [split $text (]
          set text [join $s \{]
          set s [split $text )]
          set text [join $s \}]

          return $text
      }

      default {puts stderr ":apparatus - unknown command $cmd"}
  }


}


proc :experiment {cmd args} {
###############
###############

#    :experiment command  parameters
#        command  -  join
#                    leave
#                    resources

  set path ""; set name ""


  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

# Hacked a change to disconnect to allow access to EGallow
# to allow disconnection when already disconnected
# (for double-click join when not connected)
#  SCL 11 Feb 2002
  switch $cmd {

      connect {
          set z [catch {eg "(experiment :path '$path' :name '$name' !connect)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to join experiment" $path
             error $m
          }
      }

      disconnect {
          global EGallow
          set z [catch {eg "(experiment !disconnect)"} m]
          if {$z != 0 && $m != $EGallow} then {
             display-midas-error-report self $m "Attempting to leave experiment"
             error $m
          }
      }

      new {
          set z [catch {eg "(experiment :path '$path' :name '$name' !new)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to create experiment" $path
             error $m
          }
      }

      delete {
          set z [catch {eg "(experiment :path '$path' :name '$name' !delete)"} m]
          if {$z != 0 && $m != 0x10006} then {
             display-midas-error-report self $m "Attempting to delete experiment" $path
             error $m
          }
      }

      resources {
          set z [catch {eg "(experiment ?resources)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read resources"
             error $m
          }
          return $m
      }

      default {puts stderr ":experiment - unknown command $cmd"}
  }


}

proc :add-filetype {l t} {
   set ll {}
   foreach f $l {lappend ll [list $f $t]}
   return $ll
}

proc :get-files-and-directories {path} {
   global debug
   set dl [list]; set dz 0
   set fl [list]; set fz 0

   set z [catch {glob -directory $path -types d -tails *} m]
   if {$z != 0} then {
      if $debug {insert-debug ":get-files-and-directories $path, No sub-directories found in directory"}
      set dz $z
#      display-midas-error-report self $m "Attempting to read directory" $path
#      error $m
   } else {
      set dl [:add-filetype $m "d"]
   }

   set z [catch {glob -directory $path -types f -tails *} m]
   if {$z != 0} then {
      if $debug {insert-debug ":get-files-and-directories $path, No files found in directory"}
      set fz $z
#      display-midas-error-report self $m "Attempting to read directory" $path
#      error $m
   } else {
      set fl [:add-filetype $m "-"]
   }

   if {$dz && $fz} {
      display-midas-error-report self $m "Attempting to read directory" $path
      error $m
   }

   return [concat $fl $dl]
}

proc :directory {cmd args} {
###############
###############

#    :directory command  parameters
#        command  -  files

#    changed to use glob command for Windows 7
#    extended to use -types glob parameter to distinguish between files & dirs

  set path ""; set name ""


  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

  switch $cmd {
      files {
          set z [catch {:get-files-and-directories $path} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      names {
          set z [catch {glob -directory $path -tails *} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      default {puts stderr ":directory - unknown command $cmd"}
  }

}

proc :Xdirectory {cmd args} {
###############
###############

#    :directory command  parameters
#        command  -  files

#    changed to use glob command for Windows 7

#    replaced by above 5 Sep 2012


  set path ""; set name ""


  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

  switch $cmd {
      files {
          set z [catch {glob -directory $path -tails *} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      names {
          set z [catch {glob -directory $path -tails *} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      default {puts stderr ":directory - unknown command $cmd"}
  }

}

proc :XXdirectory {cmd args} {
###############
###############

#    :directory command  parameters
#        command  -  files

#    replaced by above 3 Apr 2012

  set path ""; set name ""


  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

  switch $cmd {
      files {
          set z [catch {eg "(directory :path '$path' ?files)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      names {
          set z [catch {eg "(directory :path '$path' ?names)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to read directory" $path
             error $m
          }
          return $m
      }

      default {puts stderr ":directory - unknown command $cmd"}
  }


}

proc :spectra {cmd args} {
###############
###############

#    :spectra command  parameters
#        command  -  auth

  set user ""; set pass ""


  set i 0
  while {$i < [llength $args]} {
    set p [string range [lindex $args $i] 1 end]
    set v [lindex $args [expr $i + 1]]
    set $p $v
    incr i 2
  }

  switch $cmd {

      auth {
          set z [catch {eg "(spectrum :user '$user' :pass '$pass' !auth)"} m]
          if {$z != 0} then {
             display-midas-error-report self $m "Attempting to authorise spectrum access"
             error $m
          }
          return $m
      }

      default {puts stderr ":spectra - unknown command $cmd"}
  }


}



#    file name mapping
#      this is to handle problems caused by the same file (normally accessed
#      via nfs) having a different pathname on different systems

proc midas-map-filename {key f} {
   global env

#    global default - return asis!
   set result $f

   switch $key {
      toLynxOS -
      toVxWorks -
      toUnix  {
         if {$env(platform) == "windows"} then {
#    first try for a NFS mounted system using the drive letter M:
             set n [file split $result]
             if {[first $n] == "M:/"} then {
                 set n [lreplace $n 0 0 MIDAS]
                 set result [eval file join / $n]
             }
#    next look for a (probable) local file in  drive:/Midas or drive:/MIDAS
             set n [lrange [file split $result] 1 end]
             if {[first $n] == "Midas" || [first $n] == "MIDAS"} then {
                 set n [lreplace $n 0 0 MIDAS]
                 set result [eval file join / $n]
             }
#             insert-debug "filename $f mapped to $result"
         }
      }
      fromWindows  {
         if {$env(platform) == "unix"} then {
#    first try for a NFS mounted system using the drive letter M:
             set n [file split $result]
             if {[first $n] == "M:/" || [first $n] == "M:"} then {
                 set n [lreplace $n 0 0 MIDAS]
                 set result [eval file join / $n]
             }
#    next look for a (probable) local file in  drive:/Midas or drive:/MIDAS
             set n [lrange [file split $result] 1 end]
             if {[first $n] == "Midas" || [first $n] == "MIDAS"} then {
                 set n [lreplace $n 0 0 MIDAS]
                 set result [eval file join / $n]
             }
#             insert-debug "filename $f mapped to $result"
         }
      }
      fromUnix  {
         if {$env(platform) == "windows"} then {
#    first try for a  file system using /MIDAS
             set n [file split $result]
             if {[first $n] == "/" && [second $n] == "MIDAS"} then {
                 set n [lreplace $n 0 1 M:/]
                 set result [eval file join / $n]
             }
#             insert-debug "filename $f mapped to $result"
         }
      }
      default {}
   }
   return $result
}


#    file name conversion
#      handles problems with Windows style  c:\name\name   caused by the
#      special meaning of \ to TCL

proc midas-convert-filename {d} {
   global env

#    reprocess directory supplied to remove any \s in the string

   set f [split $d /\\]

   switch $env(platform) {
      windows {
         if {[string length [first $f]] == 2 && [string range [first $f] 1 1] == ":"} then \
            {set f [lreplace $f 0 0 "[first $f]/"]}
      }
      default {
      }
   }

   if {[string length [first $f]] == 0 && [string length $d] > 0} then {set f [lreplace $f 0 0 "/"]}

   set path ""
   foreach element $f {set path [file join $path $element]}

   return $path
}
   

proc Get_Server_OS {resource} {
    global ServerOS EGallow 

#    obtain OS in the server - this many fail 
   set EGallow 0x10004
   set ServerOS [EG "register :name '/$resource/NetVar_ServerOS' ?value"]
   set EGallow -1
}

#    file name mapping
#      this is to handle problems caused by the same file (normally accessed
#      via nfs) having a different pathname on different systems

proc Midas-Map-Filename {f} {
   global env ServerOS resource

   Get_Server_OS $resource

#    global default - return asis!
   set result $f

#    first switch on our OS then switch on server OS

  switch $env(platform) {
      windows {
          switch $ServerOS {
              Windows_NT {
                  if {[string compare -length [string length $env(MIDASAPPLBASE)] $env(MIDASAPPLBASE) $f] == 0} then {
                      set result [string range $f [string length $env(MIDASAPPLBASE)] end]
                  }
              }
              default -
              Unix -
              VxWorks -
              LynxOS {
#    first try for a NFS mounted system using the drive letter M:
                 set n [file split $result]
                 if {[first $n] == "M:/"} then {set n [lreplace $n 0 0 MIDAS]; set result [eval file join / $n]}
#    next look for a (probable) local file in  drive:/Midas or drive:/MIDAS
                 set n [lrange [file split $result] 1 end]
                 if {[first $n] == "Midas" || [first $n] == "MIDAS"} then {
                     set n [lreplace $n 0 0 MIDAS]
                     set result [eval file join / $n]
                 }
              }
          }
      }
      unix {
          switch $ServerOS {
              Windows_NT {
                  if {[string compare -length [string length $env(MIDASAPPLBASE)] $env(MIDASAPPLBASE) $f] == 0} then {
                      set result [string range $f [expr 1 + [string length $env(MIDASAPPLBASE)]] end]
#                      set result [join [file split $result] \\]
                  }
              }
              default -
              Unix -
              VxWorks -
              LynxOS {
              }
          }
      }
      default {}
  }

#  insert-debug "filename $f mapped to $result"
   return $result
}

#    load original Server Access (eg command) support

set EGreport 0
set EGallow  0xffffffff
set EGallowAll 0x0
set EGRetry 0
set EGRetries 0
set EGcommand ""
proc EGOnce {x} {
   global env tkskiperror EGreport EGallow EGallowAll EGRetry EGcommand
   global errorCode errorInfo
   if {$EGreport >= 3} then {insert-debug "executing: $x"}
   if {$EGreport >= 2} then {set-footer "executing: $x"}
   if {$EGreport == 1} then {set-footer "accessing [third $x]"}
   if {$EGreport >= 0} then {frame-busy}
   set NoError 0
   set z [catch {eg "($x)"} m]
   if {$z != 0 && $m == "" && $errorCode != "NONE"} then {set m $errorCode}

   if {$EGreport >= 3} then {insert-debug "reply: $z $m"}
   if {$z != 0 && $EGallow != $EGallowAll} then {
      if {![string is integer $m]} then {
        insert-debug "EGOnce '$x' error return: $m"
        display-midas-error-report self "Please see debug output. Malformed error return" "executing: $x"
      } else {
        if {$m != $EGallow} then {
          switch -glob -- $m {
            0x1001c -
            0x*ffff {
                set EGcommand $x
                set EGServer [first $EGcommand]
                set EGserver [second [split [third $EGcommand] /]]
                if {$EGRetry == 0} then {
                    midas-warning "Network timeout executing the [first $EGServer] server access \
command\n\n$x\n\n\nThe server $EGserver could not be contacted.\nIf the problem continues check \
the network path to the server and check the server itself.\nIf the server has crashed you should restart it."
                } else {
                    set NoError 1
                }
            }
            default {display-midas-error-report self $m "executing: $x"}
          }
          set tkskiperror 1
        }
      }
      if {$EGreport >= 0} then {frame-idle}
      if {$EGreport >= 1} then {clear-footer}
      if {$m != $EGallow && $NoError == 0} then {error $m}
   }
   if {$EGreport >= 0} then {frame-idle}
   if {$EGreport >= 1} then {clear-footer}
   return "$m"
}

proc EG {x} {
    global EGRetry EGRetries
    set EGRetry [expr $EGRetries + 1]
    while {$EGRetry > 0} {
        incr EGRetry -1
        set m [EGOnce $x]
    }
    return "$m"
}

#    load required version of EG command

if {$mytaskname != "root"} {

    if {[info exists env(ServerAccessVersion)]} then {
        set ServerAccessVersion $env(ServerAccessVersion)
    } else {
        set ServerAccessVersion 1
    }

    if {$ServerAccessVersion == 1} then {package ifneeded ServerAccess 2 [list loadsource ServerAccessV2]}
    if {$ServerAccessVersion > 1} then {loadsource ServerAccessV$ServerAccessVersion}
}


#    special bindings which may be envoked

proc EGbindings {} {

  bind all <Control-Alt-KeyPress-0> "global EGreport; set EGreport 0"
  bind all <Control-Alt-KeyPress-1> "global EGreport; set EGreport 1"
  bind all <Control-Alt-KeyPress-2> "global EGreport; set EGreport 2"
  bind all <Control-Alt-KeyPress-3> "global EGreport; set EGreport 3"
}

#    default register access procedures  -  V1 

proc write-register       {r v} {global RegisterServer; EG "register :name    '/$RegisterServer/$r' :value '$v'"}
proc write-registers      {r v} {global RegisterServer; EG "register :pattern '/$RegisterServer/$r' :value '$v'"}
proc read-register        {r}   {global RegisterServer; EG "register :name    '/$RegisterServer/$r' ?value"}
proc read-registers       {r}   {global RegisterServer; EG "register :pattern '/$RegisterServer/$r' ?names-and-values"}
proc write-attributes     {r v} {global RegisterServer; EG "register :name    '/$RegisterServer/$r' :object '$v'"}
proc read-attributes      {r}   {global RegisterServer; EG "register :name    '/$RegisterServer/$r' ?object"}
proc initialise-register  {r}   {global RegisterServer; EG "register :name    '/$RegisterServer/$r' !initialise"}
proc initialise-registers {r}   {global RegisterServer; EG "register :pattern '/$RegisterServer/$r' !initialise"}
proc define-register {class name attributes} {
   global RegisterServer

   set register /$RegisterServer/$name
   EG "register :pattern '$register' !delete"
   EG "register :name '$register' :class '$class' !new"
   EG "register :name '$register' :object '$attributes'"
}
